from zoneinfo import ZoneInfo

from strats import Clock, Strats
from strats.monitor import CronMonitor


async def cron_job(clock, state):
    print("job: ", clock.datetime)


def create_app():
    return Strats(
        clock=Clock(
            start_at="2025-01-01 12:04:50",
            tz=ZoneInfo("Asia/Tokyo"),
        ),
        monitors=[
            CronMonitor(
                cron_job=cron_job,
                cron_schedule="* * * * *",  # every minutes
            )
        ],
    ).create_app()



from dataclasses import dataclass

from strats.internal.lru_set import LRUSet


@dataclass(frozen=True)
class A:
    a: int
    b: list[int]


def test_lru_set():
    lruset = LRUSet(capacity=2)

    lruset.add(A(a=0, b=[]))
    assert lruset.contains(A(a=0, b=[]))
    assert not lruset.contains(A(a=1, b=[]))
    assert not lruset.contains(A(a=2, b=[]))

    lruset.add(A(a=0, b=[]))
    assert lruset.contains(A(a=0, b=[]))
    assert not lruset.contains(A(a=1, b=[]))
    assert not lruset.contains(A(a=2, b=[]))

    lruset.add(A(a=1, b=[]))
    assert lruset.contains(A(a=0, b=[]))
    assert lruset.contains(A(a=1, b=[]))
    assert not lruset.contains(A(a=2, b=[]))

    lruset.add(A(a=2, b=[]))
    assert not lruset.contains(A(a=0, b=[]))
    assert lruset.contains(A(a=1, b=[]))
    assert lruset.contains(A(a=2, b=[]))


def test_lru_set_with_array():
    lruset = LRUSet(capacity=2)

    lruset.add([A(a=0, b=[])])
    assert lruset.contains([A(a=0, b=[])])
    assert not lruset.contains([A(a=1, b=[])])
    assert not lruset.contains([A(a=2, b=[])])



import asyncio
from collections.abc import AsyncGenerator

from strats import Strats
from strats.monitor import StreamClient, StreamMonitor


class SampleStreamClient(StreamClient):
    async def stream(self) -> AsyncGenerator[int]:
        for i in range(10):
            await asyncio.sleep(1)
            yield i


def create_app():
    stream_monitor = StreamMonitor(client=SampleStreamClient())
    return Strats(monitors=[stream_monitor]).create_app()



import pytest
from httpx import ASGITransport, AsyncClient

from .main import create_app


@pytest.mark.asyncio
async def test_01_minimal():
    app = create_app()
    async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
        # >> healthz, metrics
        r = await client.get("/healthz")
        assert r.status_code == 200
        assert r.json() == "ok"

        r = await client.get("/metrics")
        assert r.status_code == 200

        # >> strategy
        r = await client.get("/strategy")
        assert r.status_code == 200
        assert r.json() == {"is_configured": False, "is_running": False}

        r = await client.post("/strategy/start")
        assert r.status_code == 400
        assert r.json() == {"detail": "Missing strategy configuration"}

        r = await client.post("/strategy/stop")
        assert r.status_code == 400
        assert r.json() == {"detail": "Missing strategy configuration"}

        # >> monitors
        r = await client.get("/monitors")
        assert r.status_code == 200
        assert r.json() == {"is_configured": False}

        r = await client.post("/monitors/start")
        assert r.status_code == 400
        assert r.json() == {"detail": "Missing monitors configuration"}

        r = await client.post("/monitors/stop")
        assert r.status_code == 400
        assert r.json() == {"detail": "Missing monitors configuration"}

        # >> clock
        r = await client.get("/clock")
        assert r.status_code == 200
        assert r.json()["is_real"]

        r = await client.post("/clock/start")
        assert r.status_code == 400
        assert r.json() == {"detail": "Clock is not mock"}

        r = await client.post("/clock/stop")
        assert r.status_code == 400
        assert r.json() == {"detail": "Clock is not mock"}



from dataclasses import dataclass
from unittest.mock import ANY, Mock, call

from strats.core import Data, State


@dataclass
class DummySource:
    s: int = 0


@dataclass
class DummyData:
    d: str = "0"


class DummyMetrics:
    def __init__(self, m=0.0):
        self.m: float = m


def test_data_update():
    def dummy_source_to_data(source: DummySource, _current_data: DummyData) -> DummyData:
        return DummyData(d=str(source.s))

    def dummy_data_to_metrics(data: DummyData, metrics: DummyMetrics):
        metrics.m = float(data.d)

    class DummyState(State):
        num = Data(
            data=DummyData(),
            metrics=DummyMetrics(),
            source_to_data=dummy_source_to_data,
            data_to_metrics=dummy_data_to_metrics,
        )

    state = DummyState()

    assert state.num.d == "0"
    assert DummyState.num._metrics.m == 0.0

    state.num = DummySource(s=1)
    assert state.num.d == "1"
    assert DummyState.num._metrics.m == 1.0

    del state.num
    assert state.num.d == "0"
    # assert DummyState.num._metrics.m == 0.0  # FIXME: the metrics is not reset


def test_data_lifecycle_hook():
    # Mock Parent
    lifecycle = Mock()

    # Mock children
    on_init = lifecycle.on_init
    on_delete = lifecycle.on_delete
    on_pre_event = lifecycle.on_pre_event
    on_post_event = lifecycle.on_post_event

    source_to_data = lifecycle.source_to_data
    source_to_data.side_effect = lambda s, _: DummyData(d=str(s.s))

    data_to_metrics = lifecycle.data_to_metrics

    class DummyState(State):
        num = Data(
            data=DummyData(),
            metrics=DummyMetrics(),
            source_to_data=source_to_data,
            data_to_metrics=data_to_metrics,
            on_init=on_init,
            on_delete=on_delete,
            on_pre_event=on_pre_event,
            on_post_event=on_post_event,
        )

    state = DummyState()

    # Update state data
    source = DummySource(s=123)
    state.num = source

    # Delete state data
    del state.num

    # Test the calls order
    assert lifecycle.mock_calls == [
        call.on_init(),
        call.on_pre_event(state, source),
        call.source_to_data(source, ANY),
        call.data_to_metrics(DummyData(d="123"), ANY),
        call.on_post_event(state, DummyData(d="123")),
        call.on_delete(state),
    ]



from dataclasses import dataclass
from typing import Any, Callable, Optional


@dataclass
class QueueMsg:
    source: Optional[Any] = None
    data: Optional[Any] = None
    dedup: bool = True


class Data:
    def __init__(
        self,
        *,
        data=None,
        metrics=None,
        source_to_data=None,
        data_to_metrics=None,
        on_init: Optional[Callable] = None,
        on_delete: Optional[Callable] = None,
        on_pre_event: Optional[Callable] = None,
        on_post_event: Optional[Callable] = None,
        enqueue: bool = True,
        dedup: bool = True,
    ):
        # Data and Metrics
        self.initial_data = data
        self.initial_metrics = metrics
        self._data = data
        self._metrics = metrics

        # ETL
        self.source_to_data = source_to_data
        self.data_to_metrics = data_to_metrics

        # Lifecycle Hook
        self.on_init = on_init
        self.on_delete = on_delete
        self.on_pre_event = on_pre_event
        self.on_post_event = on_post_event

        # Queue settings
        self.enqueue = enqueue
        self.dedup = dedup

        # Descriptor instance name
        self._data_name = None

    def __set_name__(self, owner, name):
        self._data_name = name

        if self.on_init is not None:
            self.on_init()

    def __get__(self, instance, owner=None):
        # When accessed as a class attribute
        if instance is None:
            return self
        return self._data

    def __set__(self, instance, new_source):
        if self.on_pre_event is not None:
            self.on_pre_event(instance, new_source)

        # Source -> Data
        if self.source_to_data is not None:
            new_data = self.source_to_data(new_source, self._data)
        else:
            new_data = new_source

        # Update Data
        self._data = new_data

        if self.enqueue and hasattr(instance, "sync_queue"):
            instance.sync_queue.put(
                QueueMsg(
                    source=new_source,
                    data=new_data,
                    dedup=self.dedup,
                )
            )

        # Update Metrics if needed
        if self.data_to_metrics is not None:
            # Data -> Metrics
            self.data_to_metrics(new_data, self._metrics)

        if self.on_post_event is not None:
            self.on_post_event(instance, new_data)

    def __delete__(self, instance):
        if self.on_delete is not None:
            self.on_delete(instance)

        # Reset data and metrics
        self._data = self.initial_data
        self._metrics = self.initial_metrics  # FIXME: metrics is not reset



class LRUSet:
    """
    A set-like data structure with Least Recently Used (LRU) eviction policy.

    This class maintains insertion order based on recency and enforces uniqueness
    using equality comparison (`==`), not hashing. Unlike standard sets, it does not
    require elements to be hashable—only that they implement `__eq__`.

    Internally, items are stored in a list in LRU order (oldest at index 0).
    Both `add` and `contains` operations have O(n) time complexity.

    When the capacity is exceeded, the least recently used item is evicted.
    """

    def __init__(self, capacity: int):
        self.capacity = capacity
        self.data: list = []  # Maintain LRU order: oldest at index 0

    def add(self, value):
        # Remove existing value (if any)
        for i, v in enumerate(self.data):
            if v == value:
                self.data.pop(i)
                break
        # Append to end (most recent)
        self.data.append(value)
        # Trim if over capacity
        if len(self.data) > self.capacity:
            self.data.pop(0)

    def contains(self, value) -> bool:
        return any(v == value for v in self.data)



from zoneinfo import ZoneInfo

from strats import Clock, Strats


def create_app():
    return Strats(
        clock=Clock(
            start_at="2025-01-01 12:00:00",
            tz=ZoneInfo("Asia/Tokyo"),
            speed=2,
        ),
    ).create_app()



from strats import Strats, StratsConfig


def create_app():
    config = StratsConfig(
        install_access_log=True,
        drop_access_log_paths=("/healthz", "/livez", "/readyz"),
    )
    return Strats(config=config).create_app()



from fastapi import FastAPI

from strats.core.kernel import Kernel

from .middleware import AccessLogMiddleware
from .router import get_kernel, router

BANNER = r"""
 _______ _______  ______ _______ _______ _______
 |______    |    |_____/ |_____|    |    |______
 ______|    |    |     \ |     |    |    ______|
"""


def kernel_getter_factory(kernel):
    def kernel_getter():
        return kernel

    return kernel_getter


class Strats(Kernel):
    def create_app(self) -> FastAPI:
        app = FastAPI()
        app.include_router(router)
        app.dependency_overrides[get_kernel] = kernel_getter_factory(self)

        if self.config.install_access_log:
            app.add_middleware(
                AccessLogMiddleware,
                drop_paths=self.config.drop_access_log_paths,
            )
        return app



import asyncio
import logging
from abc import ABC, abstractmethod
from typing import Optional

from strats.core import Clock, Monitor, State

logger = logging.getLogger(__name__)


class StreamClient(ABC):
    @abstractmethod
    async def stream(self):
        pass


class StreamMonitor(Monitor):
    def __init__(
        self,
        client: StreamClient,
        **kwargs,
    ):
        self.client = client
        self.client_name = client.__class__.__name__

        super().__init__(**kwargs)

    async def run(self, clock: Clock, state: Optional[State]):
        await self.delay()
        self.set_descriptor(state)
        logger.info(f"{self.name} start")

        success = await self.exec_on_init(clock, state)
        if not success:
            return

        try:
            async for source in self.client.stream():
                await self.exec_on_pre_event(source)
                self.update_data_descriptor(state, source)
                await self.exec_on_post_event(source)

        except asyncio.CancelledError:
            # To avoid "ERROR:asyncio:Task exception was never retrieved",
            # Re-raise the CancelledError
            raise
        except Exception as e:
            # Unexpected error
            logger.error(
                f"Stream error in {self.name}, but maybe in the `stream` function"
                f" in {self.client_name}: {e}"
            )
        finally:
            await self.exec_on_delete(clock, state)
            logger.info(f"{self.name} stopped")



import logging

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "format": "[%(asctime)s] %(levelname)s : %(name)s : %(message)s",
        },
    },
    "handlers": {
        "default": {
            "level": "INFO",
            "formatter": "default",
            "class": "logging.StreamHandler",
        },
    },
    "loggers": {
        "": {  # root logger
            "handlers": ["default"],
            "level": "INFO",
            "propagate": False,
        },
        "uvicorn": {
            "handlers": ["default"],
            "level": "INFO",
            "propagate": False,
        },
        "uvicorn.access": {
            "handlers": ["default"],
            "level": "INFO",
            "propagate": False,
        },
    },
}


class MetricsFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        return "/metrics" not in record.getMessage()



import logging
import time

from starlette.types import ASGIApp, Receive, Scope, Send

_access = logging.getLogger("strats.access")


class AccessLogMiddleware:
    def __init__(self, app: ASGIApp, drop_paths=("/healthz", "/livez", "/readyz")):
        self.app = app
        self.drop_paths = set(drop_paths)

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        if scope["type"] != "http":
            return await self.app(scope, receive, send)

        path = scope.get("path", "")
        if path in self.drop_paths:
            return await self.app(scope, receive, send)

        method = scope.get("method", "-")
        client = scope.get("client")
        addr = client[0] if client else "-"
        start = time.perf_counter()
        status_holder = {"code": 0}

        async def send_wrapped(message):
            if message["type"] == "http.response.start":
                status_holder["code"] = message.get("status", 0)
            await send(message)

        await self.app(scope, receive, send_wrapped)
        dur_ms = (time.perf_counter() - start) * 1000.0

        _access.info('%s "%s %s" %d %.2fms', addr, method, path, status_holder["code"], dur_ms)



from .cron_monitor import CronMonitor as CronMonitor
from .static_monitor import StaticMonitor as StaticMonitor
from .stream_monitor import StreamClient as StreamClient
from .stream_monitor import StreamMonitor as StreamMonitor



import asyncio
import logging
import threading
from dataclasses import dataclass, field
from typing import Optional

from prometheus_client import REGISTRY, CollectorRegistry

from .clock import Clock
from .monitor import Monitor
from .state import State
from .strategy import Strategy

logger = logging.getLogger(__name__)


@dataclass
class StratsConfig:
    install_access_log: bool = False
    drop_access_log_paths: tuple[str, ...] = field(default_factory=tuple)


class Kernel:
    def __init__(
        self,
        *,
        config: Optional[StratsConfig] = None,
        state: Optional[State] = None,
        strategy: Optional[Strategy] = None,
        monitors: Optional[list[Monitor]] = None,
        clock: Clock = Clock(),
        registry: Optional[CollectorRegistry] = None,
    ):
        self.config = config or StratsConfig()
        self.state = state
        self.state_stop_event = None

        # There is no event loop yet, so don't create an `asyncio.Event`.
        self.monitors = monitors
        self.monitor_tasks: dict[str, asyncio.Task] = {}
        self.monitor_started_ats: dict[str, asyncio.Task] = {}

        self.strategy = strategy
        self.strategy_task = None
        self.strategy_started_at = None

        self.clock = clock
        self.clock_task = None

        self.registry = registry or REGISTRY

    async def start_strategy(self):
        if self.strategy is None:
            raise ValueError("Missing strategy configuration")

        if self.state is not None:
            self.state.set_queues()

        if self.strategy_task and not self.strategy_task.done():
            return

        # Since only Strategy reads from the queue, the queue is prepared in start_strategy.
        if self.state is not None:
            self.state.flush_queue()

        self.strategy_task = asyncio.create_task(
            _handle_error(self.strategy.run)(self.clock, self.state),
            name="strategy",
        )
        self.strategy_started_at = self.clock.datetime

    async def stop_strategy(self, timeout=5.0):
        if self.strategy is None:
            raise ValueError("Missing strategy configuration")
        if self.strategy_task.done():
            raise ValueError("Strategy is already stopped")

        if "__str__" in type(self.strategy).__dict__:
            logger.info("stop strategy")
            logger.info(f"strategy details: {self.strategy}")

        self.strategy_started_at = None
        self.strategy_task.cancel()
        try:
            await self.strategy_task
        except asyncio.CancelledError:
            pass

    async def start_monitors(self):
        if self.monitors is None:
            raise ValueError("Missing monitors configuration")

        if self.state is not None:
            self.state.set_queues()

            self.state_stop_event = threading.Event()
            self.state.run(self.state_stop_event)

        for monitor in self.monitors:
            task = self.monitor_tasks.get(monitor.name)
            if task and not task.done():
                continue

            self.monitor_tasks[monitor.name] = asyncio.create_task(
                _handle_error(monitor.run)(self.clock, self.state),
                name=monitor.name,
            )
            self.monitor_started_ats[monitor.name] = self.clock.datetime

    async def stop_monitors(self, timeout=5.0):
        if self.monitors is None:
            raise ValueError("Missing monitors configuration")

        self.monitor_started_ats = {}

        for task in self.monitor_tasks.values():
            if not task.done():
                # we should await canceled task complete
                # cf. https://docs.python.org/ja/3.13/library/asyncio-task.html#asyncio.Task.cancel
                task.cancel()
                try:
                    await task
                except asyncio.CancelledError:
                    pass

        if self.state is not None:
            self.state_stop_event.set()
            self.state.sync_to_async_queue_thread.join()

    async def start_clock(self):
        if not self.clock.is_mock:
            raise ValueError("Clock is not mock")

        if self.clock_task and not self.clock_task.done():
            return

        self.clock_task = asyncio.create_task(
            _handle_error(self.clock.run)(),
            name="clock",
        )

    async def stop_clock(self, timeout=5.0):
        if not self.clock.is_mock:
            raise ValueError("Clock is not mock")
        if self.clock_task.done():
            raise ValueError("Clock is already stopped")

        self.clock_task.cancel()
        try:
            await self.clock_task
        except asyncio.CancelledError:
            pass


def _handle_error(func):
    async def wrapper(*args, **kwargs):
        try:
            await func(*args, **kwargs)
        except asyncio.CancelledError:
            logger.info("handle cancel process")
            raise
        except Exception as e:
            logger.error(f"got unexpected exception: {e}")

    return wrapper



import typer
import uvicorn

app = typer.Typer(no_args_is_help=True)


@app.command()
def runserver(
    target: str = typer.Option("main:app", "--target"),
    host: str = "0.0.0.0",
    port: int = 8000,
    reload: bool = True,
    access_log: bool = False,
    factory: bool = typer.Option(
        False,
        "--factory",
        help="Treat target as an application factory, not an ASGI app instance",
    ),
):
    """Run development server (default: main:app)."""
    # target をそのまま uvicorn.run に渡す
    uvicorn.run(
        target,
        host=host,
        port=port,
        reload=reload,
        access_log=access_log,
        log_config=None,
        factory=factory,
    )



import asyncio
import logging
from typing import Callable, Optional

from strats.core import Clock, Monitor, State

logger = logging.getLogger(__name__)


class StaticMonitor(Monitor):
    def __init__(
        self,
        static_job: Callable,
        **kwargs,
    ):
        self.static_job = static_job

        super().__init__(**kwargs)

    async def run(self, clock: Clock, state: Optional[State]):
        await self.delay()
        self.set_descriptor(state)
        logger.info(f"{self.name} start")

        success = await self.exec_on_init(clock, state)
        if not success:
            return

        try:
            await self.static_job(clock, state)

        except asyncio.CancelledError:
            # To avoid "ERROR:asyncio:Task exception was never retrieved",
            # Re-raise the CancelledError
            raise
        except Exception as e:
            # Unexpected error
            logger.error(f"Error in {self.name}: {e}")
        finally:
            await self.exec_on_delete(clock, state)
            logger.info(f"{self.name} stopped")



from dataclasses import dataclass
from decimal import Decimal
from typing import Optional

from prometheus_client import REGISTRY, CollectorRegistry, Counter, Gauge


@dataclass
class PricesData:
    bid: Decimal = Decimal("0")
    ask: Decimal = Decimal("0")


class PricesMetrics:
    def __init__(
        self,
        *,
        registry: Optional[CollectorRegistry] = None,
        namespace: Optional[str] = None,
        subsystem: Optional[str] = None,
    ) -> None:
        reg = registry or REGISTRY

        self.bid = Gauge(
            "prices_bid",
            "Bid price.",
            namespace=namespace or "",
            subsystem=subsystem or "",
            registry=reg,
        )
        self.ask = Gauge(
            "prices_ask",
            "Ask price.",
            namespace=namespace or "",
            subsystem=subsystem or "",
            registry=reg,
        )
        self.spread = Gauge(
            "prices_spread",
            "Ask - Bid.",
            namespace=namespace or "",
            subsystem=subsystem or "",
            registry=reg,
        )
        self.update_count_total = Counter(
            "prices_update_count_total",
            "Number of price updates.",
            namespace=namespace or "",
            subsystem=subsystem or "",
            registry=reg,
        )


def prices_data_to_prices_metrics(data: PricesData, metrics: PricesMetrics):
    metrics.bid.set(float(data.bid))
    metrics.ask.set(float(data.ask))
    metrics.spread.set(float(data.ask - data.bid))
    metrics.update_count_total.inc()



import asyncio

import pytest
from httpx import ASGITransport, AsyncClient

from .main import create_app


@pytest.mark.asyncio
async def test_02_stream_monitor():
    app = create_app()
    async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
        # >> healthz, metrics
        res = await client.get("/healthz")
        assert res.status_code == 200
        assert res.json() == "ok"

        res = await client.get("/metrics")
        assert res.status_code == 200

        # >> strategy
        res = await client.get("/strategy")
        expect = {"is_configured": False, "is_running": False}
        assert res.status_code == 200
        assert res.json() == expect

        res = await client.post("/strategy/start")
        expect = {"detail": "Missing strategy configuration"}
        assert res.status_code == 400
        assert res.json() == expect

        res = await client.post("/strategy/stop")
        expect = {"detail": "Missing strategy configuration"}
        assert res.status_code == 400
        assert res.json() == expect

        # >> monitors
        res = await client.get("/monitors")
        expect = {
            "is_configured": True,
            "monitors": {
                "StreamMonitor_1": {
                    "is_running": False,
                },
            },
        }
        assert res.status_code == 200
        assert res.json() == expect

        res = await client.post("/monitors/start")
        expect = {
            "is_configured": True,
            "monitors": {
                "StreamMonitor_1": {
                    "is_running": True,
                },
            },
        }
        assert res.status_code == 200
        resjson = res.json()
        # started_at は可変のため比較から除外
        resjson["monitors"]["StreamMonitor_1"].pop("started_at", None)
        assert resjson == expect

        # 監視ループが進むのを少し待つ
        await asyncio.sleep(0.5)

        res = await client.post("/monitors/stop")
        expect = {
            "is_configured": True,
            "monitors": {
                "StreamMonitor_1": {
                    "is_running": False,
                },
            },
        }
        assert res.status_code == 200
        assert res.json() == expect



import asyncio
import logging
import re

import pytest
from httpx import ASGITransport, AsyncClient

from .main import create_app


@pytest.mark.asyncio
async def test_04_strategy_state_stream_monitor(caplog):
    app = create_app()
    async with AsyncClient(transport=ASGITransport(app=app), base_url="http://test") as client:
        # >> healthz, metrics
        res = await client.get("/healthz")
        assert res.status_code == 200
        assert res.json() == "ok"

        res = await client.get("/metrics")
        assert res.status_code == 200

        # >> strategy
        res = await client.get("/strategy")
        expect = {"is_configured": True, "is_running": False}
        assert res.status_code == 200
        assert res.json() == expect

        # >> monitors
        res = await client.get("/monitors")
        expect = {
            "is_configured": True,
            "monitors": {"StreamMonitor_1": {"is_running": False}},
        }
        assert res.status_code == 200
        assert res.json() == expect

        # >> run
        res = await client.post("/strategy/start")
        expect = {"is_configured": True, "is_running": True}
        assert res.status_code == 200
        resjson = res.json()
        resjson.pop("started_at", None)  # 可変のため除外
        assert resjson == expect

        res = await client.post("/monitors/start")
        expect = {
            "is_configured": True,
            "monitors": {"StreamMonitor_1": {"is_running": True}},
        }
        assert res.status_code == 200
        resjson = res.json()
        resjson["monitors"]["StreamMonitor_1"].pop("started_at", None)  # 可変のため除外
        assert resjson == expect

        # ログ検証に備えて INFO を捕捉
        with caplog.at_level(logging.INFO):
            # 監視・ストラテジが 1 チック動くのを待つ
            await asyncio.sleep(0.5)

        # >> check
        res = await client.get("/metrics")
        assert res.status_code == 200
        body = res.text
        assert extract_unlabeled_metric_value(body, "prices_bid") == 100.0
        assert extract_unlabeled_metric_value(body, "prices_ask") == 101.0
        assert extract_unlabeled_metric_value(body, "prices_spread") == 1.0
        assert extract_unlabeled_metric_value(body, "prices_update_count_total") == 1.0

        # stderr 代わりに caplog でログメッセージを確認
        all_msgs = "\n".join(caplog.messages)
        assert "strategy > bid: 100" in all_msgs

        # >> stop
        res = await client.post("/monitors/stop")
        expect = {
            "is_configured": True,
            "monitors": {"StreamMonitor_1": {"is_running": False}},
        }
        assert res.status_code == 200
        assert res.json() == expect

        res = await client.post("/strategy/stop")
        expect = {"is_configured": True, "is_running": False}
        assert res.status_code == 200
        assert res.json() == expect


def extract_unlabeled_metric_value(body: str, metric_name: str) -> float:
    pattern = rf"^{re.escape(metric_name)} ([0-9.e+-]+)"
    match = re.search(pattern, body, re.MULTILINE)
    if not match:
        raise ValueError(f"Metric {metric_name} not found")
    return float(match.group(1))
